#!/usr/bin/env python
#
# PkgTrackInst.PlistWriter - Writes pkg-plist based on install log
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42, (c) Poul-Henning Kamp):
# Hye-Shik Chang <perky@FreeBSD.org> wrote this file.  As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some
# day, and you think this stuff is worth it, you can buy me a beer in return.
#
# Hye-Shik Chang
# ----------------------------------------------------------------------------
# $Id: PlistWriter.py 644 2006-06-25 09:44:16Z perky $

__pychecker__ = 'no-argsused no-isliteral'

import os, sys
import time, re
from cStringIO import StringIO
from PkgTrackInst import options
from PkgTrackInst.cmd_support import errx, echo_msg
from PkgTrackInst.compat23 import *
from PkgTrackInst.Trackers import InstFile, InstDirectory, TrackRoot

SUBST_IGNORE = ('PREFIX', 'PORTDOCS', 'LOCALBASE', 'X11BASE')
MANPAGELIST_FILENAME = 'MANPAGES_LIST'
MANPAGELIST_LINEWIDTH = 64


class PlistWriter(object):

    def __init__(self):
        self.plist_subst = list()
        self.prefix = self.get_make_env('PREFIX')
        if options.substitute:
            self.build_sub_list()

    def get_output(self, cmd):
        f = os.popen(cmd, 'r')
        r = f.read()
        status = f.close()
        if status:
            errx("command `%s' exited with error code %d" % (cmd, status))
        return r

    def get_make_env(self, varname):
        return self.get_output('make -V ' + varname + ' ' +
                               ' '.join(options.makeargs)).strip()

    def build_sub_list(self):
        pat_subst = re.compile('([A-Za-z_0-9]+)=("([^"]*)"|[^"]\S*)')
        substlist = self.get_make_env('PLIST_SUB')

        for var, raw, stripped in pat_subst.findall(substlist):
            if var not in SUBST_IGNORE:
                self.plist_subst.append((stripped or raw, var))
        self.plist_subst.sort(reverse=True)

    def write(self):
        if os.path.exists(options.output):
            os.rename(options.output, options.output + '.bak')

        echo_msg('Generating %s' % options.output)
        self.clear_manpath()
        self.generate_plist(open(options.output, 'w'))

    def relpath(self, fullpath):
        prefix = self.prefix.split(os.sep)
        full = fullpath.split(os.sep)
        if full[:len(prefix)] == prefix:
            repled = os.sep.join(full[len(prefix):])
            for orig, subst in self.plist_subst:
                if repled.startswith(orig+'/') or repled == orig:
                    return '%%' + subst + '%%' + repled[len(orig):]
            return repled
        else:
            return fullpath

    def clear_manpath(self):
        pat_manfile = re.compile('([A-Za-z_0-9.-]+)\.([0-9ln])(\.gz)?')
        manpages = {}

        for category in '123456789ln':
            pages = []
            for ftype in ('man', 'cat'):
                path = os.sep.join((self.prefix, 'man', ftype+category))

                node = TrackRoot[path]
                if not isinstance(node, InstDirectory):
                    errx("`%s' is not a directory." % path)

                for name, child in node.iterchildren():
                    if isinstance(child, InstDirectory):
                        continue # ignore subdirectories in man path.

                    m = pat_manfile.match(name)
                    if not m:
                        continue # ignore files with non-uniform name

                    m_subject, m_category, m_suffix = m.groups()
                    if category != m_category:
                        errx("`%s' is miscategorized in the installation" %
                             path)
                    pages.append(m_subject)
                    child.created = False

            if pages:
                manpages[category] = pages

        if manpages:
            echo_msg("Writing manpages list as `%s'" % MANPAGELIST_FILENAME)

            f = self.open_manlist_output()
            for cate, pages in manpages.iteritems():
                out = ['']
                for page in sorted(pages):
                    item = page + '.' + cate
                    if (len(item) + len(out[-1]) > MANPAGELIST_LINEWIDTH and
                            out[-1]): # put 1 at least
                        out[-1] += ' \\'
                        out.append('')
                    out[-1] += ' ' + item

                print >> f, 'MAN%s=\t%s' % (cate.upper(), out[0][1:])
                for line in out[1:]:
                    print >> f, '\t%s' % line[1:]

    def open_manlist_output(self):
        return open(MANPAGELIST_FILENAME, 'w')

    def generate_plist(self, out, relsub=True):
        fileents, dirents = list(), list()
        def gather_entries(ent):
            if ent.created:
                if isinstance(ent, InstDirectory):
                    dirents.append(ent)
                else:
                    fileents.append(ent)
        TrackRoot.walk(gather_entries)

        fileents.sort(key=lambda x: x.abspath)
        dirents.sort(key=lambda x: x.abspath)

        if relsub:
            translatepath = self.relpath
        else:
            translatepath = lambda x: x

        for ent in fileents:
            print >> out, translatepath(ent.abspath)
        for ent in reversed(dirents):
            print >> out, '@dirrm ' + translatepath(ent.abspath)


class PlistChecker(PlistWriter):

    def __init__(self):
        PlistWriter.__init__(self)
        self.pkgdb = os.path.join(self.get_make_env('PKG_DBDIR'),
                                  self.get_make_env('PKGNAME'))
        self.checkresult = None
        self.tweak_flags = {}
        if options.output == 'pkg-plist': # default
            options.output = None

    def open_output(self):
        if not hasattr(self, 'output'):
            if options.output:
                self.output = open(options.output, 'w')
            else:
                self.output = sys.stderr

    def check_plist(self):
        prefix = '/'
        dbfiles, dbdirs = set(), set()
        for line in open(os.path.join(self.pkgdb, '+CONTENTS')):
            if line.startswith('@cwd'):
                prefix = line.split()[1]
            elif line.startswith('@dirrm'):
                cmd, dirpath = line.split(None, 1)
                dbdirs.add(os.path.join(prefix, dirpath.strip()))
            elif line.startswith('@unexec'):
                if 'gconftool-2 --makefile-uninstall-rule' in line:
                    self.tweak_flags['ignore_gconf_schema'] = True
            elif line.startswith('@'):
                continue # ignore another commands
            else:
                dbfiles.add(os.path.join(prefix, line.strip()))

        # GCONF_SCHEMAS invokes gconftool uninstall using @unexec.
        if self.tweak_flags.get('ignore_gconf_schema'):
            del TrackRoot[os.path.join(self.prefix,
                            'etc/gconf/gconf.xml.defaults')]

        plist = StringIO()
        self.generate_plist(plist, relsub=False)
        plist.seek(0, 0)

        realfiles, realdirs = set(), set()
        for line in plist:
            if line.startswith('@dirrm'):
                cmd, dirpath = line.split(None, 1)
                realdirs.add(dirpath.strip())
            elif line.startswith('@'):
                continue
            else:
                realfiles.add(line.strip())

        self.checkresult = (realfiles, realdirs, dbfiles, dbdirs)

    def write(self):
        self.check_plist()
        self.print_check_result()

    def open_manlist_output(self):
        return self.manlist

    def print_check_result(self):
        if not self.checkresult:
            errx("Plist is not checked yet.")

        realfiles, realdirs, dbfiles, dbdirs = self.checkresult

        missingfiles = realfiles - dbfiles
        if missingfiles:
            self.print_listing(missingfiles,
                    "Files installed but not mentioned in pkg-plist",
                    isdir=False, relsub=True)

        missingdirs = realdirs - dbdirs
        if missingdirs:
            self.print_listing(missingdirs,
                    "Directories created but not mentioned in pkg-plist",
                    isdir=True, relsub=True)

        obsoletefiles = dbfiles - realfiles
        if obsoletefiles:
            self.print_listing(obsoletefiles,
                    "Files mentioned in pkg-plist but not installed",
                    isdir=False, relsub=False)

        obsoletedirs = dbdirs - realdirs
        if obsoletedirs:
            self.print_listing(obsoletedirs,
                    "Directories mentioned in pkg-plist but not created",
                    isdir=True, relsub=False)

        all = sum(map(len, (missingfiles, missingdirs,
                            obsoletefiles, obsoletedirs)))
        if all < 1:
            echo_msg("Your pkg-plist is perfect for now!")
        else:
            echo_msg("Found %d error(s) in your pkg-plist." % all)

    def print_listing(self, content, title, isdir, relsub):
        self.open_output()

        print >> self.output
        print >> self.output, "=== %s ===" % title

        for entry in sorted(content, reverse=isdir):
            if relsub:
                outname = self.relpath(entry)
            else:
                outname = entry

            if isdir:
                outname = '@dirrm ' + outname
            print >> self.output, outname
